#pragma once

#include <vc_base.hpp>

#include <vdmonitors.h>

#include <graphics_sink.hpp>


/// an app-sharing virtual channel.
class appshare_virtual_channel : public vcsdk::virtual_channel_base {
private:

	overlay_vc_controller* _overlay_controller; /// the virtual channel controller.
	std::unique_ptr<vcsdk::graphics_sink> _prov; /// the graphics sink which will provide app-sharing or screen-sharing image frames.
	std::string _vc_name; /// the virtual channel name.

public:

	/// constructor.
	appshare_virtual_channel(USHORT usChannelNumber, const std::string& name, overlay_vc_controller* controller)
		: vcsdk::virtual_channel_base(usChannelNumber, controller)
		, _overlay_controller(controller)
		, _vc_name(name) {
	}

	/// writes some shared data for the virtual channel controller.
	void writeSharedData(LPBYTE lpData, USHORT& usLen) override {
		usLen = sizeof(VDOVERLAY_C2H);
	}

	/// process incoming data for this virtual channel.
	INT ondata(LPBYTE pBuf, USHORT Length) override {
		appshare_message* pMsg = (appshare_message*)pBuf;

		switch (pMsg->command) {
		case start_app_sharing:
			on_start_sharing();
			break;

		case stop_app_sharing:
			on_stop_sharing();
			break;

		case subscribe_app:
			on_subscribe(GfxSurf_Application, pMsg->app_id, pMsg->overlay_id);
			break;

		case unsubscribe_app:
			on_unsubscribe(GfxSurf_Application, pMsg->app_id, pMsg->overlay_id);
			break;

		case subscribe_screen:
			on_subscribe(GfxSurf_Monitor, pMsg->app_id, pMsg->overlay_id);
			break;

		case unsubscribe_screen:
			on_unsubscribe(GfxSurf_Monitor, pMsg->app_id, pMsg->overlay_id);
			break;

		default:
			break;
		}

		return CLIENT_STATUS_SUCCESS;
	}

	/// handle display changes for this virtual channel.
	void onviewportchanged() override {
		//ToDo: unsubscribe monitors and re-subscribe
	}

	/// authorize a graphics listener, if this virtual channel has one.
	bool authorize_graphics_listener(GBufferConnectionEvt* pEvt) override {
		return _prov && _prov->authorize(pEvt);
	};

	/// enable app-sharing or screen-sharing by creating a graphics sink. called when the 'start_app_sharing' command is received from the virtual desktop.
	void on_start_sharing() {
		_prov.reset(new vcsdk::graphics_sink(_controller->handle(), _vc_name, [this](GfxEvent evt, const void* data) {
			on_graphics_sink_event(evt, data);
			}));
	}

	/// release the graphics sink and reset the overlays. called when the 'stop_app_sharing' command is received from the virtual desktop.
	void on_stop_sharing() {
		_prov.reset();
		for (auto& win : _overlay_controller->get()) { win.second.reset_image(); }
	}

	/// begin an app-sharing or screen-sharing session. called whenever a 'subscribe_app' or 'subscribe_screen' command is received from the virtual desktop.
	void on_subscribe(const GfxSurfaceType& type, unsigned int id, unsigned int overlay_id) {
		if (_prov) {
			_prov->add_capture_session(id, type, overlay_id, [this, overlay_id](GfxEvent evt, const void* data) {
				on_capture_session_event(overlay_id, evt, data);
				});
		}
	}

	/// terminate an app-sharing or screen-sharing session. called whenever a 'unsubscribe_app' or 'unsubscribe_screen' command is received from the virtual desktop.
	void on_unsubscribe(const GfxSurfaceType& type, unsigned int id, unsigned int overlay_id) {
		if (_prov) {
			_prov->delete_capture_session(id, type, overlay_id);
		}
	}

	/// process an event from the graphics sink.
	void on_graphics_sink_event(GfxEvent evt, const void* data) {
		switch (evt) {
		case GfxEvt_Connected:
			on_connected();
			break;
		}
	}

	/// performs various graphics sink post-connection duties. called when the graphics sink is connected to the graphics service/provider.
	void on_connected() {
		on_app_sharing_started();
		update_screens();
	}

	/// send a status message to the virtual desktop application so that it can enable app-sharing and screen-sharing functionality.
	void on_app_sharing_started() {
		appshare_message msg{};
		msg.command = app_sharing_status;
		msg.status = app_sharing_started;

		MEMORY_SECTION buf = {};
		buf.pSection = (LPBYTE)&msg;
		buf.length = sizeof(msg);

		write(buf);
	}

	/// send a list of screens to the virtual desktop application so that it can display/present them to the user.
	void update_screens() {
		VIRTUAL_MONITOR_LAYOUT* pInfo{};
		if (VdGetMonitorInfo(_controller->handle(), &pInfo) == CLIENT_STATUS_SUCCESS) {
			appshare_message msg{};
			msg.command = appsh_command_ids::update_screens;
			msg.num_screens = pInfo->NumMonitors;

			MEMORY_SECTION buf{};
			buf.length = sizeof(msg) + (sizeof(unsigned) * pInfo->NumMonitors);
			buf.pSection = new BYTE[buf.length];

			unsigned int* ptr = (unsigned int*)(buf.pSection + sizeof(msg));
			std::transform(&pInfo->Monitors[0], &pInfo->Monitors[pInfo->NumMonitors], &ptr[0], [](const VIRTUAL_MONITOR& monitor) { return monitor.Id; });
			memcpy(buf.pSection, &msg, sizeof(msg));

			write(buf);

			delete[]buf.pSection;
		}
	}

	/// process an event for a capture session.
	void on_capture_session_event(unsigned int overlay_id, GfxEvent evt, const void* data) {
		switch (evt) {
		case GfxEvt_Subscribed:
			on_capture_session_started(overlay_id, *reinterpret_cast<const GfxSurfaceDescriptor*>(data));
			break;

		case GfxEvt_Unsubscribed:
			on_capture_session_stopped(overlay_id, *reinterpret_cast<const GfxSurfaceDescriptor*>(data));
			break;

		case GfxEvt_FrameUpdate:
			on_capture_frame(overlay_id, *reinterpret_cast<const GfxImage*>(data));
			break;
		}
	}

	void on_capture_session_started(unsigned int overlay_id, const GfxSurfaceDescriptor& subscription) {
		//ToDo: inform virtual desktop. mostly informative.
		//ToDo: post to message loop
	}

	void on_capture_session_stopped(unsigned int overlay_id, const GfxSurfaceDescriptor& subscription) {
		//ToDo: inform virtual desktop.
		//ToDo: post to message loop
	}

	/// process an image frame for an app-sharing or screen-sharing session. presents or displays the image in the configured overlay window.
	void on_capture_frame(unsigned int overlay_id, const GfxImage& image) {
		//ToDo: post to message loop???
		_overlay_controller->with_overlay(overlay_id, [&image](overlay_window& win) {
			win.reset_image(bitmap::from_pixels(win.get_device_context(), { static_cast<long>(image.width), static_cast<long>(image.height) }, image.data));
			});
	}
};