#pragma once

#include "overlay_window.hpp"
#include "util.hpp"

#include <vc_base.hpp>
#include <vdversion.h>

#include <map>


class overlay_virtual_channel : public vcsdk::virtual_channel_base {
private:

	overlay_vc_controller* _overlay_controller;

	void send_response(unsigned int id, RECT rc) {
		window_message msg{};
		msg.command = reposition_window;
		msg.id = id;
		msg.pos.x = rc.left;
		msg.pos.y = rc.top;
		msg.size.dx = rc.right - rc.left;
		msg.size.dy = rc.bottom - rc.top;

		MEMORY_SECTION buf{};
		buf.pSection = (LPBYTE)&msg;
		buf.length = sizeof(msg);

		write(buf);
	}

public:

	/// constructor.
	overlay_virtual_channel(USHORT usChannelNumber, const std::string& name, overlay_vc_controller* controller)
		: vcsdk::virtual_channel_base(usChannelNumber, controller)
		, _overlay_controller(controller) {
	}

	/// writes some shared data for the virtual channel controller.
	void writeSharedData(LPBYTE lpData, USHORT& usLen) override {

		usLen = sizeof(VDOVERLAY_C2H);

		if (lpData != NULL) {

			// initialize default data
			PVDOVERLAY_C2H pVdData = reinterpret_cast<PVDOVERLAY_C2H>(lpData);
			pVdData->Overlay.id = 1;
			pVdData->Overlay.pos.x = 50;
			pVdData->Overlay.pos.y = 50;
			pVdData->Overlay.size.dx = 600;
			pVdData->Overlay.size.dy = 400;

			PVDCWAVERSION pCWAVer = &pVdData->CWAVersion;
			(void)VdGetICAVersion(NULL, &pCWAVer->major, &pCWAVer->minor, &pCWAVer->revision, &pCWAVer->build);

			if (bitmap img = bitmap::from_resource(_overlay_controller->appInstance(), IDB_BACKGROUND); img.valid()) {
				SIZE res = img.dimensions();
				pVdData->Overlay.size.dx = res.cx;
				pVdData->Overlay.size.dy = res.cy;
			}
		}
	}

	/// process incoming data for this virtual channel.
	INT ondata(LPBYTE pBuf, USHORT Length) override {
		window_message* pMsg = (window_message*)pBuf;

		switch (pMsg->command) {
		case create_window:
			_overlay_controller->add(pMsg->id, L"the overlay");
			break;

		case destroy_window:
			_overlay_controller->remove(pMsg->id);
			break;

		case reparent_window:
			(void)_overlay_controller->with_overlay(pMsg->id, [&](overlay_window& win) { return win.reparent((HWND)pMsg->parent); });
			break;

		case reposition_window:
			(void)_overlay_controller->with_overlay(pMsg->id, [&](overlay_window& win) {
				RECT rc = {};
				win.get_window_rect(&rc);
				send_response(pMsg->id, rc);
				return win.reposition(RECT_from_position_and_size(pMsg->pos, pMsg->size));
				});
			break;

		case update_clipping:
			(void)_overlay_controller->with_overlay(pMsg->id, [&](overlay_window& win) {

				std::vector<RECT> rects;
				rectangle* clipRect = (rectangle*)(pMsg + 1);

				std::transform(clipRect, &clipRect[pMsg->num_clip_rects], std::back_inserter(rects), RECT_from_rectangle);

				return win.clip(rects);
				});
			break;

		default:
			break;
		}

		return 0;
	}

	/// handle display changes for this virtual channel.
	void onviewportchanged() override {
		for (auto& win : _overlay_controller->get()) {

			win.second.post_message(WM_USER_VIEWPORT_CHANGED, 0, 0);

#if 0 // will cause deadlock
			// recalculate parent from original parent
			win.second.reparent(win.second.get_original_parent());

			// recalculate position from original position
			win.second.reposition(win.second.get_original_position());

			// recalculate clipping rectangles from original clipping rectangles
			win.second.clip(win.second.get_original_clip_rects());
#endif
		}
	}

};