#pragma once

#include <managed_window.hpp>
#include <image_button.hpp>
#include <static_control.hpp>
#include <slider_control.hpp>

#include "app_preview.hpp"
#include "screen_sharing.hpp"

#include <sstream>
#include <map>
#include <variant>
#include <algorithm>


class app_sharing_tray : public managed_window {
private:

	/// app preview image control
	class preview : public image_button<1> {
	private:

		// note: prefer variants to raw pointers so that we don't have to handle
		//       null pointer cases everywhere the member is accessed.

		using content_type = std::variant<application_preview, screen_content>; /// polymorphic content type
		content_type _source;

	public:

		static const int height = 100;
		static const int width = 100;

		/// constructor. construct with underlying window handle, bitmap image and application window.
		template <typename T>
		preview(HWND hwnd, T&& source)
			: image_button(hwnd, { source.make_preview_image(hwnd, {width, height}) })
			, _source(std::move(source)) {
		}

		/// copy constructor. copy construction is not allowed.
		preview(const preview&) = delete;

		/// move constructor.
		preview(preview&&) = default;

		/// destructor.
		~preview() {
			destroy();
		}

		template <typename T>
		static preview create(const window& parent, int nID, RECT rc, T&& source) {
			return preview(image_button::create(parent, nID, rc), std::move(source));
		}

		/// retrieve the application window.
		content_type& source() { return _source; }

		template <typename Fn>
		auto using_content(Fn&& fn) {
			return std::visit(std::forward<Fn>(fn), _source);
		}
	};



	static const int control_id = 11000;

	static_control _label;
	slider_control _screens_switch; /// slider control for toggling app-sharing and screen-sharing mode.
	std::map<std::size_t, preview> _apps; /// the application/screen preview image controls.
	std::vector<unsigned int> _screen_ids; /// a saved copy of the screen ids received from the local endpoint device

	wfvirtualchannel _vc;

public:

	/// constructor. default constructor.
	app_sharing_tray()
		: managed_window(NULL, null_deleter) {
	}

	/// constructor.
	app_sharing_tray(HWND hwnd, DeleterFn&& deleter, const wfvirtualchannel& overlay_vc)
		: managed_window(hwnd, std::move(deleter))
		, _label(static_control::create(hwnd, control_id - 2, L"Screens", {0, 0, 100, 20}))
		, _screens_switch(slider_control::create(hwnd, control_id - 1, {0, 0, 20, 20}))
		, _vc(overlay_vc) {

		enable_horizontal_scrollbar();
		refresh();
		position_controls();
	}

	/// move constructor.
	app_sharing_tray(app_sharing_tray&&) = default;

	/// move assignment operator.
	app_sharing_tray& operator =(app_sharing_tray&&) = default;

	/// main window factory method
	static app_sharing_tray create(HINSTANCE hInstance, const std::wstring& title, const wfvirtualchannel& overlay_vc) {
		using window_class_type = window_class<app_sharing_tray>;
		static window_class_type wndclass = window_class_type::register_class(hInstance, L"ctxoverlay_appshare_window_class", (HBRUSH)(COLOR_WINDOW + 0));

		return managed_window::create(wndclass, title, null_deleter, overlay_vc);
	}

	/// retrieve the displayed height of the app-sharing tray.
	unsigned int get_height() const {
		if (visible()) {
			return preview::height + _screens_switch.get_dimensions().cy;
		}

		return 0;
	}

	/// update the list of screens. called whenever a new list of screens is received from the local endpoint device.
	void update_screens(unsigned int* screen_ids, unsigned int num_screens) {
		std::copy(screen_ids, screen_ids + num_screens, std::back_inserter(_screen_ids));
	}

	/// called whenever the window position changes. moves the remote overlays.
	void on_position_changed() {
		std::ranges::for_each(_apps, [this](auto& app) {
			RECT rc{};
			app.second.client_rect_to_screen(&rc);
			app.second.using_content([&rc](auto& content) {
				content.target().reposition(rc.left, rc.top);
				content.target().resize(rc.right - rc.left, rc.bottom - rc.top);
				});
			});
	}

	/// called on some timer interval. recalculates overlay clipping.
	void do_clipping(clip_region& clip) {
		std::ranges::for_each(_apps, [this, &clip](auto& app) {
			app.second.using_content([&app, &clip](auto& content) {
				content.target().clip(app.second, clip);
				});
			});
	}

	void reset() {
		_apps.clear(); // release all previews and overlays.
		_screens_switch.turn_off(); // reset the screens toggle.
	}

	void refresh() {
		refresh(_screens_switch.is_on());
	}

private:

	void refresh(bool screens) {
		if (screens) {
			refresh_screens();
		}
		else {
			refresh_app_list();
		}
	}

	template <typename T>
	static auto to_lparam(std::vector<T>& vec) { return reinterpret_cast<LPARAM>(&vec); }

	template <typename T>
	static auto& from_lparam(LPARAM lparam) { return *reinterpret_cast<std::vector<T>*>(lparam); }

	/// refresh the application list. generates new previews for the new list of applications enumerated.
	void refresh_app_list() {

		std::vector<HWND> wins;
		EnumWindows([](HWND hwnd, LPARAM lparam) {
			from_lparam<HWND>(lparam).push_back(hwnd);
			return TRUE;
			}, to_lparam(wins));

		auto it = std::partition(wins.begin(), wins.end(), [&](HWND hwnd) {
			window win(hwnd);
			return !win.get_owner().valid()
				&& !(win.get_extended_style() & WS_EX_TOOLWINDOW)
				&& !(win.get_extended_style() & WS_EX_LAYERED)
				&& !(win.get_style() & WS_POPUP)
				&& (win.get_style() & WS_VISIBLE)
//              && (win.get_extended_style() & WS_EX_APPWINDOW)
//				&& hwnd != _hwnd // filter out ourself
				&& hwnd != get_parent()
				;
			});

		wins.erase(it, wins.end());

		load_previews(wins, [](HWND win) { return application_preview(win); });
	}

	/// refresh the screen previews. generates new previews for the current set of screens.
	void refresh_screens() {
		load_previews(_screen_ids, [this, n{ 10 }](const unsigned int& item) mutable { // start overlay ids at 10 because the main window creates up to 9 overlays of it's own
			return screen_content(item, get_parent(), remote_window::create(_vc, n++, (window_handle)_hwnd));
			});
	}

	template <typename T, typename Fn>
	std::pair<std::size_t, preview> load_preview(const T& item, int n, Fn&& fn) {
		const std::size_t ID = control_id + n;
		RECT rc{ preview::width * n, 0, preview::width * (n + 1), preview::height };
		return std::pair{ ID, preview::create(_hwnd, (int)ID, rc, fn(item)) };
	}

	template <typename T, typename Fn>
	void load_previews(std::vector<T>& items, Fn&& fn) {
		_apps.clear();

		// update scroll bar. depending on the number of previews we need to generate, we may need scroll bars.
		auto dims = get_dimensions();
		update_horizontal_scrollbar(static_cast<int>(items.size() * preview::width), dims.cx);

		// generate the previews
		std::ranges::transform(items, std::inserter(_apps, std::begin(_apps)), [this, n{ 0 }, fn = std::move(fn)](const T& item) mutable { return load_preview(item, n++, fn); });

		// render the preview content
		std::ranges::for_each(_apps, [this](auto& app) {
			app.second.using_content([this](auto& content) {
				content.render(_hwnd, WM_USER_SESSIONEVENT, 0);
				});
			});

		// make the controls visible
		std::ranges::for_each(_apps, [](auto& item) { item.second.show(); });

		on_position_changed();
	}

	void position_controls() {
		_label.move_window({ 40, 100, 140, 120 });
		_label.show();

		_screens_switch.move_window({ 10, 100, 30, 120 });
		_screens_switch.show();
	}

	LRESULT on_message(UINT message, WPARAM wparam, LPARAM lparam, bool& handled) {
		switch (message) {
		case WM_MOVE:
			on_position_changed();
			handled = true;
			break;

		case WM_SIZE:
			on_size_changed(LOWORD(lparam), HIWORD(lparam));
			handled = true;
			break;

		case WM_HSCROLL:
			on_hscroll(LOWORD(wparam));
			break;

		case WM_COMMAND:
			on_command(LOWORD(wparam), HIWORD(wparam));
			break;

		case WM_DRAWITEM:
			handled = true;
			on_draw_item(reinterpret_cast<LPDRAWITEMSTRUCT>(lparam));
			return 1;

		case WM_USER_SESSIONEVENT + session_event::started:
			on_session_started(lparam, wparam);
			break;

		case WM_USER_SESSIONEVENT + session_event::stopped:
			break;

		default:
			break;
		}
		return 0;
	}

	void on_size_changed(UINT32 dx, UINT32 dy) {}

	void on_draw_item(LPDRAWITEMSTRUCT lpDraw) {
		if (lpDraw->hwndItem == _screens_switch) {
			//lpDraw->hDC, lpDraw->rcItem
		}
	}

	void on_command(int nID, int notification) {
		switch (notification) {
		case BN_CLICKED:
			on_button_clicked(nID);
			break;

		default:
			break;
		}
	}

	void on_button_clicked(int nID) {
		switch (nID) {
		case control_id - 1: // slider control
			toggle_screens();
			break;

		default:
			on_item_selected(nID);
			break;
		}
	}

	void on_item_selected(int nID) {
		if (auto it = _apps.find(nID); it != _apps.end()) {

			bool is_screen_content = std::holds_alternative<screen_content>(it->second.source());

			it->second.using_content([this, is_screen_content](auto& content) {
				get_parent().send_message(WM_USER_SELECTAPP, (WPARAM)(is_screen_content ? 1 : 0), (LPARAM)content.id());
				});
		}
	}

	void toggle_screens() {
		bool was_on = _screens_switch.toggle();
		refresh(!was_on);
	}

	auto find_content_by_session_key(const std::size_t& key) {
		return std::ranges::find_if(_apps, [&key](auto& app) {
			return app.second.using_content([&key](auto& c2) {
				return c2.id() == key; });
			});
	}

	void on_session_started(const std::size_t& key, const std::size_t& id) {
		if (auto it = find_content_by_session_key(key); it != _apps.end()) {
			it->second.using_content([&id](auto& content) {
				content.on_event(session_event::started, (unsigned int)id);
				});
		}
	}

	void on_hscroll(int mode) {
		auto [page, pos, track_pos] = get_scroll_info(SB_HORZ);
		auto new_pos = pos;
		switch (mode) {
		case SB_LINELEFT:
			new_pos -= preview::width;
			break;

		case SB_LINERIGHT:
			new_pos += preview::width;
			break;

		case SB_PAGELEFT:
			new_pos -= page;
			break;

		case SB_PAGERIGHT:
			new_pos += page;
			break;

		case SB_THUMBPOSITION:
			new_pos = track_pos;
			break;

		case SB_THUMBTRACK:
			new_pos = track_pos;
			break;

		default:
			break;
		}

		set_scroll_position_x(new_pos);
		if (new_pos = get_scroll_position_x(); new_pos != pos) {
			scroll(pos - new_pos, 0);
		}
	}
};