#pragma once

#include "remote_window.hpp"
#include "control_bar.hpp"
#include "app_share_tray.hpp"
#include "app_sharing.hpp"
#include "grabber.hpp"
#include "zorder_iterator.hpp"

#include <managed_window.hpp>
#include <clip_region.hpp>
#include <dialog.hpp>

#include <vector>
#include <iterator> // for std::back_inserter
#include <algorithm> // for std::transform and others
#include <numeric> // for std::accumulate
#include <sstream>


// necessary for comparing vectors of RECTs
inline bool operator ==(const RECT& lhs, const RECT& rhs) {
    return lhs.left == rhs.left &&
        lhs.right == rhs.right &&
        lhs.top == rhs.top &&
        lhs.bottom == rhs.bottom;
}


void extend_y(SIZE& sz, LONG y) { sz.cy += y; }
void extend_y(RECT& rc, LONG y) { rc.bottom += y; }
void shrink_y(SIZE& sz, LONG y) { extend_y(sz, -y); }
void shrink_y(RECT& rc, LONG y) { extend_y(rc, -y); }


auto test_occlusion(const window& win, window winTest) {
    RECT rcRef{}; // reference rectangle
    win.client_rect_to_screen(&rcRef);

    RECT rcTest{}; // test rectangle
    winTest.get_window_rect(&rcTest);

    RECT rcOcc{};
    bool occludes = winTest.visible() && !winTest.iconic() && IntersectRect(&rcOcc, &rcRef, &rcTest);

    return std::pair(occludes, rcOcc);
}


clip_region detect_occlusions(const window& win) {

    std::vector<clip_region> rgns;
    zorder_iterator begin(win, GW_HWNDPREV);
    zorder_iterator end(NULL, GW_HWNDPREV);

    for (++begin; begin != end; ++begin) {
        auto [covers, rc] = test_occlusion(win, *begin);
        if (covers) rgns.push_back(clip_region::from_rectangle(rc));
    }

    return std::accumulate(rgns.begin(), rgns.end(), clip_region(), [](clip_region rgn, const clip_region& other) { return rgn | other; });
}

clip_region get_visible_region(const window& win) {
    clip_region sysrgn = clip_region::from_window(win);
    clip_region occlusions = detect_occlusions(win);
    return occlusions ^ sysrgn;
}





class main_window : public managed_window, public wftaskqueue {
private:

    using default_content = display_content<remote_window>;
    using content_type = std::variant<screen_content, application_content, default_content>; /// polymorphic content type

    std::vector<content_type> _remotes;
    clip_region _clip;

    wfsession _session;
    wfvirtualchannel _overlay_vc;
    wfvirtualchannel _app_share_vc;
    wfvirtualchannel::reader _overlay_vc_reader;
    wfvirtualchannel::reader _app_share_vc_reader;

    control_window _controls;
    app_sharing_tray _app_tray;
    grabber _grabber;

    /// hardcoded maximum number of tiles
    static constexpr unsigned int num_tiles = 9;

public:

    /// constructor. default constructor.
    main_window()
        : managed_window(NULL, null_deleter) {
    }

    /// constructor.
    main_window(HWND hndl, DeleterFn&& deleter, const wfsession& session, const wfvirtualchannel& overlay_vc, const wfvirtualchannel& app_share_vc)
        : managed_window(hndl, std::move(deleter))
        , _overlay_vc(overlay_vc)
        , _app_share_vc(app_share_vc)
        , _session(session)
        , _controls(control_window::create(get_app_instance(), L"", session))
        , _app_tray(app_sharing_tray::create(get_app_instance(), L"apps", overlay_vc))
        , _grabber(grabber::create(hndl, IDC_GRABBER, { 0, 0, 118, 16 })) {

        _controls.embed_into(_hwnd);
        _app_tray.embed_into(_hwnd);
        _grabber.disable();
        _grabber.show();

        reset();

        // set clipping timer. poll every half second
        SetTimer(hndl, 0, 500, NULL);

        start_appsharing();

        // re-post to same thread to start receiving data on the virtual channel. ensures window creation completes first.
        post_task(wfasynctask(allocate_queue_id("main_window startup"), [hndl]() {
            window win(hndl);
            main_window* main = (main_window*)win.get_user_data();
            main->start_receiving();
            return true;
            }));
    }

    /// constructor. move constructor.
    main_window(main_window&& other) noexcept
        : managed_window(std::move(other)) {

        swap(std::move(other));
    }

    /// move assignment operator.
    main_window& operator =(main_window&& other) noexcept {
        if (this != &other) {
            managed_window::operator=(std::move(other));
            swap(std::move(other));
        }
        return *this;
    }

    /// main window factory method
    static main_window create(HINSTANCE hInstance, const std::wstring& title, const wfsession& session, const wfvirtualchannel& overlay_vc, const wfvirtualchannel& app_share_vc) {
        using window_class_type = window_class<main_window>;
        static window_class_type wndclass = window_class_type::register_class(hInstance, L"ctxoverlay_main_window_class", CreateSolidBrush(RGB(133, 227, 229)));

        return managed_window::create(wndclass, title, post_quit_deleter, session, overlay_vc, app_share_vc);
    }

    /// destructor.
    ~main_window() {
        stop_appsharing();
    }

private:

    void swap(main_window&& other) noexcept {
        std::swap(_remotes, other._remotes);
        std::swap(_clip, other._clip);
        std::swap(_overlay_vc, other._overlay_vc);
        std::swap(_app_share_vc, other._app_share_vc);
        std::swap(_overlay_vc_reader, other._overlay_vc_reader);
        std::swap(_app_share_vc_reader, other._app_share_vc_reader);
        std::swap(_session, other._session);
        std::swap(_controls, other._controls);
        std::swap(_app_tray, other._app_tray);
        std::swap(_grabber, other._grabber);
    }



    remote_window create_overlay(unsigned int overlay_id) {
        return remote_window::create(_overlay_vc, overlay_id, (window_handle)_hwnd);
    }

    remote_window create_overlay(unsigned int overlay_id, const RECT& rc) {
        return remote_window::create(_overlay_vc, overlay_id, (window_handle)_hwnd, rc);
    }

    /// reset the application back to it's initial state.
    void reset() {

        // if we're app sharing, stop it and clear the selected app
        _app_tray.hide(); // hide the appsharing tray initially
        _app_tray.reset();

        RECT rc{};
        client_rect_to_screen(&rc);
        shrink_y(rc, _controls.get_height());
        shrink_y(rc, _app_tray.get_height());

        // clear any existing overlays and create a new overlay on the local endpoint device.
        _remotes.clear();
        _remotes.emplace_back(default_content(1, create_overlay(1, rc)));

        // resize the client area to match the native dimensions of the remote overlay to avoid having to stretch the image.
        SIZE sz = {};
        get_overlay_dimensions(&sz);
        extend_y(sz, _controls.get_height());
        extend_y(sz, _app_tray.get_height());
        resize_client_area(sz);
    }

    /// demonstrates multiple concurrent overlays and complex clipping by displaying up to 'num_tiles' overlays in a grid formation.
    void tile() {
        _remotes.clear();

        std::generate_n(std::back_inserter(_remotes), num_tiles, [this, id{ 0 }]() mutable { id++;
            return default_content(id + 1, create_overlay(id + 1));
            });

        on_size_changed(0, 0);
        on_position_changed();
    }

    void start_appsharing() {
        if (_app_share_vc) {
            appshare_message msg{};
            msg.command = start_app_sharing;

            ULONG ignore;
            _app_share_vc.write(msg, ignore);
        }
    }

    void stop_appsharing() {
        if (_app_share_vc) {
            appshare_message msg{};
            msg.command = stop_app_sharing;

            ULONG ignore;
            _app_share_vc.write(msg, ignore);
        }
    }

    void start_receiving() {
        // we have to be careful of using the this pointer as it will be invalidated whenever the move constructor is invoked.
        _overlay_vc_reader = _overlay_vc.start_receiving(this, std::bind(&main_window::on_overlay_vc_data, this, std::placeholders::_1));
        if (_app_share_vc) {
            _app_share_vc_reader = _app_share_vc.start_receiving(this, std::bind(&main_window::on_app_share_vc_data, this, std::placeholders::_1));
        }
    }

    std::size_t allocate_queue_id(const std::string& label) override {
        return WM_USER_VCDATA;
    }

    void post_task(wfasynctask&& task) override {

        // move the task to the heap so that it can be passed to another thread.
        std::unique_ptr<wfasynctask> async_task(new wfasynctask(std::move(task)));

        // post task from vc reader thread to window message loop thread.
        if (post_message(static_cast<UINT>(async_task->id()), NULL, (LPARAM)async_task.get())) {

            // successfully posted the task to the main thread, so release ownership of it to prevent it from being
            // destroyed when async_task goes out of scope.
            async_task.release();
        }
    }

    /// run an asynchronous task. invoked from the main thread (window message loop).
    void execute_task(wfasynctask* async_task) {
        std::unique_ptr<wfasynctask> task(async_task); // re-attach to async task so that it is properly destroyed when we go out of scope
        task->run();
    }

    /// handle incoming data from the overlay virtual channel.
    void on_overlay_vc_data(wfbuffer& data) {
        if (window_message* msg = data.as<window_message>(); msg) {
            _controls.update_overlay_geometry(*msg);
        }
    }

    /// handle incoming data from the app sharing virtual channel.
    void on_app_share_vc_data(wfbuffer& data) {
        if (appshare_message* msg = data.as<appshare_message>(); msg) {
            switch (msg->command) {
            case app_sharing_status:
                _grabber.enable();
                break;

            case update_screens:
                _app_tray.update_screens((unsigned int*)(msg + 1), msg->num_screens);
                break;
            }
        }
    }

    /// get the native dimensions of the remote overlay image.
    void get_overlay_dimensions(SIZE* psz) {

        VDOVERLAY_C2H data = {};
        _overlay_vc.get_client_data(data);

        if (data.Overlay.size.dx != 0 &&
            data.Overlay.size.dy != 0) {

            psz->cx = data.Overlay.size.dx;
            psz->cy = data.Overlay.size.dy;
        }
    }

    template <typename Fn>
    auto using_content(content_type& remote, Fn&& fn) {
        return std::visit(std::forward<Fn>(fn), remote);
    }

    template <typename Fn>
    auto using_contents(Fn&& fn) {
        std::ranges::for_each(_remotes, [this, &fn](auto& remote) {
            using_content(remote, std::forward<Fn>(fn));
            });
    }

    /// divide an array of items into a square number of tiles.
    unsigned int compute_layout(SIZE& tileSize) {
        unsigned int sz = std::max((unsigned int)ceil(sqrt((double)_remotes.size())), (unsigned int)1); // protect against division by 0 if vector is empty.

        RECT rc = {};
        get_client_rect(&rc);
        shrink_y(rc, _controls.get_height());
        shrink_y(rc, _app_tray.get_height());

        tileSize.cx = rc.right / sz;
        tileSize.cy = rc.bottom / sz;

        return sz;
    }

    /// compute the row and column from the tile position (index) and size (dimensions).
    template <typename T>
    std::pair<T, T> compute_row_column(const T& index, unsigned int sz) {
        return { index % sz, index / sz };
    }

    /// called whenever the window position changes. moves the remote overlays.
    void on_position_changed() {

        // account for window frame. we want to send the position of the client area in screen coordinates
        POINT pt = {};
        client_to_screen(&pt);

        SIZE tileSize = {};
        unsigned int sz = compute_layout(tileSize);

        using_contents([this, &pt, &sz, &tileSize, i{0}](auto& content) mutable {

            auto [row, col] = compute_row_column(i++, sz);

            POINT pos{};
            pos.x = pt.x + (tileSize.cx * row);
            pos.y = pt.y + (tileSize.cy * col);

            content.target().reposition(pos.x, pos.y);
            });

        _controls.send_message(WM_USER, 0, 0);
        _app_tray.on_position_changed();
    }

    /// called whenever the window dimensions change. resizes the remote overlays.
    void on_size_changed(UINT32 dx, UINT32 dy) {

        SIZE tileSize = {};
        unsigned int sz = compute_layout(tileSize);

        using_contents([&tileSize](auto& content) {
            content.target().resize(tileSize.cx, tileSize.cy);
            });

        // adjust the position of the control bar
        RECT rcControls{};
        get_client_rect(&rcControls);
        rcControls.top = rcControls.bottom - _controls.get_height();
        _controls.move_window(rcControls);

        // adjust the position of the appsharing tray
        RECT rcApps{};
        get_client_rect(&rcApps);
        rcApps.bottom -= _controls.get_height();
        rcApps.top = rcApps.bottom - _app_tray.get_height();
        _app_tray.move_window(rcApps);

        // adjust the position of the grabber
        RECT rcGrabber{};
        _grabber.get_client_rect(&rcGrabber);
        auto grabber_width = rcGrabber.right - rcGrabber.left;
        rcGrabber.top = rcApps.top - 16;
        rcGrabber.bottom = rcApps.top;
        rcGrabber.left = ((rcControls.right - rcControls.left) / 2) - (grabber_width / 2);
        rcGrabber.right = rcGrabber.left + grabber_width;
        _grabber.move_window(rcGrabber);
    }

    /// called periodically to check for window occlusions.
    void on_time() {
        clip_region new_clip = get_visible_region(*this);
        if (!EqualRgn(_clip, new_clip) && new_clip.to_rectangle_list() != _clip.to_rectangle_list()) {

            for (std::size_t i = 0; i < _remotes.size(); ++i) {

                // figure out the size of each tile based upon the number of overlays we have
                SIZE tileSize = {};
                unsigned int sz = compute_layout(tileSize);

                // compute the position of the current overlay
                auto [row, col] = compute_row_column(static_cast<unsigned int>(i), sz);

                /// compute the tile offset based on the tile size and row and column indices
                SIZE offset{};
                offset.cx = tileSize.cx * row;
                offset.cy = tileSize.cy * col;

                // compute the rectangle bounds, in screen coordinates, of the current overlay
                POINT pt{};
                client_to_screen(&pt);

                RECT rc{};
                rc.left = pt.x + offset.cx;
                rc.top = pt.y + offset.cy;
                rc.right = rc.left + tileSize.cx;
                rc.bottom = rc.top + tileSize.cy;

                // create a clip region from the rectangle bounds of the current overlay and intersect it with the new clip region.
                clip_region rgn = new_clip & clip_region::from_rectangle(rc);

                // ...remove the grabber rectangle from the visible region
                RECT rcGrabber{};
                _grabber.get_window_rect(&rcGrabber);
                rgn = rgn - clip_region::from_rectangle(rcGrabber);

                // convert the rectangle back into client area coordinates, and then translate the resulting rectangle to be relative
                // to the tile offset.
                std::vector<RECT> mapped;
                std::ranges::transform(rgn.to_rectangle_list(), std::back_inserter(mapped), [&](RECT r) {
                    MapWindowPoints(HWND_DESKTOP, _hwnd, (LPPOINT)&r, 2);
                    OffsetRect(&r, -offset.cx, -offset.cy);
                    return r;
                    });

                using_content(_remotes[i], [&mapped](auto& content) { content.target().clip(mapped); });
            }

            _app_tray.do_clipping(new_clip);

            std::swap(_clip, new_clip);
        }
    }

    void on_mouse_move(SHORT x, SHORT y) {
        std::wostringstream ss;
        ss << "Mouse: ";
        ss << x;
        ss << ", ";
        ss << y;
    }

    void on_display_changed(SHORT dx, SHORT dy, SHORT bpp) {

        _controls.send_message(WM_DISPLAYCHANGE, MAKELPARAM(dx, dy), MAKEWPARAM(bpp, 0));
        //_controls.send_message(WM_DISPLAYCHANGE, dx, dy);
        //_controls.send_message(WM_DISPLAYCHANGE, wparam, lparam);
    }

    /// called whenever the user right-clicks within the content area of the window. displays a context menu to the user.
    void on_context_menu(int x, int y) {
        HMENU hMenu = LoadMenu(NULL, MAKEINTRESOURCE(IDR_POPUPMENU));
        TrackPopupMenu(GetSubMenu(hMenu, 0), TPM_RIGHTBUTTON, x, y, 0, _hwnd, NULL);
    }

    /// called whenever the user clicks on the grabber control. hides or shows the app-sharing tray.
    void toggle_app_tray() {
        RECT rc{};
        get_window_rect(&rc);

        if (_app_tray.visible()) {
            _app_tray.hide();
            _app_tray.reset();
            _grabber.set_state(grabber::closed);
            shrink_y(rc, 100); // shrink the window to reclaim space occupied by the app sharing tray
        } else {
            _app_tray.refresh();
            _app_tray.show();
            _grabber.set_state(grabber::open);
            extend_y(rc, 100); // expand the window to make room for the app sharing tray
        }

        move_window(rc);
    }

    void on_grabber(int notification) {
        switch (notification) {
        case BN_CLICKED:
            toggle_app_tray();
            break;
        }
    }

    template <typename Fn>
    void replace_content(Fn&& fn) {
#if 1
        std::vector<content_type> new_remotes;
        std::ranges::transform(_remotes, std::back_inserter(new_remotes), [fn = std::move(fn)](content_type& content) {
            return std::visit([fn = std::move(fn)](auto& value) -> content_type {
                return { fn(std::move(value.target())) }; // steal the overlay window from the old content using move semantics
                }, content);
            });

        std::swap(_remotes, new_remotes);

        using_contents([this](auto& content) {
            content.render(_hwnd, WM_USER_SESSIONEVENT, WFAppDisplayBorder | WFAppDisplayCursor);
            });
#else
        // here's an alternative implementation that doesn't modify the container values by stealing the overlay. c++ standard
        // indicates that the transform function should not modify values of the container.
        RECT rc{};
        client_rect_to_screen(&rc);
        shrink_y(rc, _controls.get_height());
        shrink_y(rc, _app_tray.get_height());

        _remotes.clear();
        _remotes.emplace_back({ fn(create_overlay(1, rc)) });
#endif
    }

    /// called whenever the selects an item from the app-sharing tray. displays the app-sharing or screen-sharing content.
    void on_select_app(bool screen, LPARAM lparam) {
        if (screen) {
            replace_content([this, &lparam](remote_window&& overlay) {
                return screen_content((UINT)lparam, _hwnd, std::move(overlay));
                });
        }
        else {
            replace_content([this, &lparam](remote_window&& overlay) {
                return application_content(reinterpret_cast<HWND>(lparam), _hwnd, std::move(overlay));
                });
        }
    }

    /// called whenever a button or menu item is clicked/selected.
    void on_command(int id, int notification) {
        switch (id) {
        case IDM_CLOSE:
            destroy();
            break;

        case IDM_RESET:
            reset();
            break;

        case IDM_TILE:
            tile();
            break;

        case IDM_HIDE:
            break;

        case IDC_GRABBER:
            on_grabber(notification);
            break;

        case IDM_ABOUT:
            DialogBoxParam(get_app_instance(), MAKEINTRESOURCE(IDD_ABOUTBOX), _hwnd, About, (LPARAM)this);
            break;
        }
    }

    /// dispatches/handles window messages/events.
    LRESULT on_message(UINT message, WPARAM wparam, LPARAM lparam, bool& handled) {
        switch (message) {
        case WM_MOVE:
            on_position_changed();
            handled = true;
            break;

        case WM_SIZE:
            on_size_changed(LOWORD(lparam), HIWORD(lparam));
            handled = true;
            break;

        case WM_TIMER:
            on_time();
            break;

        case WM_CONTEXTMENU:
            on_context_menu(GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam));
            break;

        case WM_COMMAND:
            on_command(LOWORD(wparam), HIWORD(wparam));
            break;

        case WM_DISPLAYCHANGE:
            on_display_changed(LOWORD(lparam), HIWORD(lparam), LOWORD(wparam));
            break;

        case WM_MOUSEMOVE:
            on_mouse_move(GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam));
            break;

        case WM_CLOSE:
            on_close();
            break;

        case WM_USER_VCDATA:
            execute_task(reinterpret_cast<wfasynctask*>(lparam));
            break;

        case WM_USER_SELECTAPP:
            on_select_app(wparam == 1, lparam);
            break;

        case WM_USER_SUBSCRIBEAPP:
            on_subscribe_app(wparam & 0x1, static_cast<unsigned int>(lparam), static_cast<unsigned int>(wparam >> 1));
            break;

        case WM_USER_SUBSCRIBESCREEN:
            on_subscribe_screen(wparam & 0x1, static_cast<unsigned int>(lparam), static_cast<unsigned int>(wparam >> 1));
            break;

        case WM_USER_SESSIONEVENT:
            on_session_event(lparam, wparam);
            break;

        default:
            break;
        }

        return 0;
    }

    auto find_content_by_session_key(const std::size_t& key) {
        return std::ranges::find_if(_remotes, [this, &key](auto& remote) {
            return using_content(remote, [&key](auto& content) {
                return content.id() == key; });
            });
    }

    void on_session_event(const std::size_t& key, const std::size_t& id) {
        if (auto it = find_content_by_session_key(key); it != _remotes.end()) {
            using_content(*it, [&id](auto& content) {
                content.on_event(session_event::started, (unsigned int)id);
                });
        }
    }

    void on_close() {
        // the application exit process involves three events: WM_CLOSE, WM_DESTROY, and WM_QUIT.
        // We're stopping the current app-sharing on WM_CLOSE, because when WM_DESTROY is processed,
        // the window handle is set to NULL and the class destructor is invoked. The app_display
        // class sends messages to this class using the window handle, which would be NULL in this
        // class' destructor. WM_QUIT exits the message loop and allows the application to terminate.
        //_app.reset();
        _remotes.clear();
    }

    void on_subscribe_app(bool subscribe, unsigned int bufferid, unsigned int overlay_id) {
        appshare_message msg{};
        msg.command = subscribe ? subscribe_app : unsubscribe_app;
        msg.app_id = bufferid;
        msg.overlay_id = overlay_id;

        ULONG ignore;
        _app_share_vc.write(msg, ignore);
    }

    void on_subscribe_screen(bool subscribe, unsigned int bufferid, unsigned int overlay_id) { // bufferid = monitorid
        appshare_message msg{};
        msg.command = subscribe ? subscribe_screen : unsubscribe_screen;
        msg.app_id = bufferid;
        msg.overlay_id = overlay_id;

        ULONG ignore;
        _app_share_vc.write(msg, ignore);
    }

    std::wstring get_client_name() {
        if (auto info = _session.query_client_info(); info.has_value()) {
            return info->name();
        }
        return L"client name unknown";
    }

    std::wstring get_client_version() {
        if (VDOVERLAY_C2H info = {}; _overlay_vc.get_client_data(info)) {
            std::wostringstream ss;
            ss << info.CWAVersion.major << ".";
            ss << info.CWAVersion.minor << ".";
            ss << info.CWAVersion.revision << ".";
            ss << info.CWAVersion.build;
            return ss.str();
        }

        return L"client version unknown";
    }

    std::wstring get_product_id() {
        if (auto info = _session.query_client_info(); info.has_value()) {
            std::wostringstream ss;
            ss << "CWA For ";
            ss << to_string(info->productId());
            return ss.str();
        }

        return L"client product id unknown";
    }

    int select_cwa_icon() {
        int icon = -1;
        if (auto info = _session.query_client_info(); info.has_value()) {
            switch (info->productId()) {
            case wfcpi_windows:
                icon = IDI_WINDOWS_ICON;
                break;

            case wfcpi_linux:
                icon = IDI_LINUX_ICON;
                break;

            case wfcpi_mac:
                icon = IDI_MAC_ICON;
                break;

            case wfcpi_ios:
                icon = IDI_IOS_ICON;
                break;

            case wfcpi_android:
                icon = IDI_ANDROID_ICON;
                break;

            default:
                break;
            }
        }

#ifdef _DEBUG
        return IDI_WINDOWS_ICON;
#else
        return icon;
#endif
    }

    /// message handler for about box.
    static INT_PTR CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
        switch (message) {
        case WM_INITDIALOG: {
            main_window* win = reinterpret_cast<main_window*>(lParam);
            HICON hIcon = LoadIcon(win->get_app_instance(), MAKEINTRESOURCE(win->select_cwa_icon()));

            dialog dlg(hDlg);
            dlg.get_item(IDC_CLIENT_PRODUCT_ID).set_text(win->get_product_id());
            dlg.get_item(IDC_CLIENT_VERSION).set_text(win->get_client_version());
            dlg.get_item(IDC_CLIENT_NAME).set_text(win->get_client_name());
            dlg.get_static(IDC_CWA_ICON).set_icon(hIcon);

            return (INT_PTR)TRUE;
        }

        case WM_COMMAND:
            if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
                EndDialog(hDlg, LOWORD(wParam));
                return (INT_PTR)TRUE;
            }
            break;
        }
        return (INT_PTR)FALSE;
    }
};