#pragma once

#include <window.hpp>
#include <clip_region.hpp>

#include <vector>
#include <algorithm>
#include <iterator>


/// A remotely controlled window via Citrix session
class remote_window {
private:

	wfvirtualchannel _conn; /// the virtual channel connection
	VDOVERLAY_C2H _data; /// the virtual channel specific data
	position _pos; /// the current position of the overlay
	extents _sz; /// the current size of the overlay
	unsigned int _id;
	bool _should_destroy;

public:

	/// create a remote window with id and parent.
	static remote_window create(const wfvirtualchannel& vc, unsigned int id, window_handle parent) {
		remote_window window(vc, id);
		window.create(parent);
		return window;
	}

	/// create a remote window with id, parent, and initial position and size.
	static remote_window create(const wfvirtualchannel& vc, unsigned int id, window_handle parent, const RECT& rc) {
		remote_window window = create(vc, id, parent);
		window.reposition(rc.left, rc.top);
		window.resize(rc.right - rc.left, rc.bottom - rc.top);
		return window;
	}

	/// constructor. constructs a new remote window in the current session.
	remote_window(const wfvirtualchannel& vc, unsigned int id)
		: _conn(vc)
		, _data({})
		, _pos({})
		, _sz({})
		, _id(id)
		, _should_destroy(true) {

		// retrieve the client data
		_conn.get_client_data(_data);
	}

	/// constructor. move constructor.
	remote_window(remote_window&& other)
		: _conn(std::move(other._conn))
		, _data({})
		, _pos({})
		, _sz({})
		, _id(0)
		, _should_destroy(false) {

		std::swap(_conn, other._conn);
		std::swap(_data, other._data);
		std::swap(_pos, other._pos);
		std::swap(_sz, other._sz);
		std::swap(_id, other._id);
		std::swap(_should_destroy, other._should_destroy);
	}

	/// destructor.
	~remote_window() {
		if (_should_destroy) {
			destroy(); // inform the remote endpoint
		}
	}

	/// move assignment operator
	remote_window& operator =(remote_window&& other) {
		if (this != &other) {
			std::swap(_data, other._data);
			std::swap(_pos, other._pos);
			std::swap(_sz, other._sz);
			std::swap(_id, other._id);
			std::swap(_should_destroy, other._should_destroy);
		}
		return *this;
	}

	unsigned int id() const { return _id; }

	/// create the window on the remote endpoint.
	void create(window_handle parent) {
		window_message msg = {};
		msg.command = create_window;
		msg.id = _id;
		msg.parent = parent;
		msg.size.dx = 600;
		msg.size.dy = 400;

		ULONG written = 0;
		(void)_conn.write(msg, written);

		msg.command = reparent_window;
		(void)_conn.write(msg, written);

		msg.command = reposition_window;
		(void)_conn.write(msg, written);
	}

	/// destroy the window on the remote endpoint.
	void destroy() {
		window_message msg = {};
		msg.command = destroy_window;
		msg.id = _id;

		ULONG written = 0;
		(void)_conn.write(msg, written);
	}

	/// update the window position on the remote endpoint
	void reposition(unsigned int x, unsigned int y) {

		if (_pos.x != x ||
			_pos.y != y) {
			_pos.x = x;
			_pos.y = y;

			window_message msg = {};
			msg.command = reposition_window;
			msg.id = _id;
			msg.size.dx = _sz.dx;
			msg.size.dy = _sz.dy;
			msg.pos.x = _pos.x;
			msg.pos.y = _pos.y;

			ULONG written = 0;
			(void)_conn.write(msg, written);
		}
	}

	/// update the window size on the remote endpoint
	void resize(unsigned int dx, unsigned int dy) {
		if (_sz.dx != dx ||
			_sz.dy != dy) {
			_sz.dx = dx;
			_sz.dy = dy;

			window_message msg = {};
			msg.command = reposition_window;
			msg.id = _id;
			msg.size.dx = _sz.dx;
			msg.size.dy = _sz.dy;
			msg.pos.x = _pos.x;
			msg.pos.y = _pos.y;

			ULONG written = 0;
			(void)_conn.write(msg, written);
		}
	}

	void clip(unsigned int count, PRECT pRects) {
		unsigned int len = sizeof(window_message) + (count * sizeof(RECT));
		LPBYTE pBuf = new BYTE[len];
		memset(pBuf, 0, len);
		window_message* pMsg = (window_message*)pBuf;
		pMsg->command = update_clipping;
		pMsg->id = _id;
		pMsg->num_clip_rects = count;
		rectangle* pr = (rectangle*)(pMsg + 1);
		for (unsigned int i = 0; i < count; ++i) {
			pr[i].tl.x = pRects[i].left;
			pr[i].tl.y = pRects[i].top;
			pr[i].br.x = pRects[i].right;
			pr[i].br.y = pRects[i].bottom;
		}
		ULONG written = 0;
		(void)_conn.write(pBuf, len, written);
		delete pBuf;
	}


	void clip(std::vector<RECT>& rectangles) {
		if (rectangles.size() > 0) {
			clip(static_cast<unsigned int>(rectangles.size()), rectangles.data());
		} else {
			RECT AllZeroesClippingRectangle{ 0, 0, 0, 0 };
			clip(1, &AllZeroesClippingRectangle);
		}
	}

	void clip(const window& win, clip_region& new_clip) {

		RECT rc{};
		win.client_rect_to_screen(&rc);

		clip_region rgn = new_clip & clip_region::from_rectangle(rc);

		std::vector<RECT> mapped;
		std::ranges::transform(rgn.to_rectangle_list(), std::back_inserter(mapped), [&](RECT r) {
			MapWindowPoints(HWND_DESKTOP, win, (LPPOINT)&r, 2);
			return r;
			});

		clip(mapped);
	}
};