#pragma once


/// a c++ iterator for enumerating windows in z-order.
class zorder_iterator {
private:

    HWND _hwnd; /// the current window in the z-order.
    int _dir;

public:

    // these type definitions are required to make the iterator compatible with std::ranges.
    // the c++ standard requires 'difference_type' even though it seems unnecessary for my
    // use case. zorder_iterator does not overload the difference operator.

    using difference_type = std::ptrdiff_t;
    using value_type = HWND;

    /// constructor.
    zorder_iterator(HWND hWnd, int dir)
        : _hwnd(hWnd)
        , _dir(dir) {
    }

    /// prefix increment. advance the iterator to the next window in the z-order.
    zorder_iterator& operator ++() {
        _hwnd = GetNextWindow(_hwnd, _dir);
        return *this;
    }

    /// postfix increment. advance the iterator to the next window in the z-order.
    zorder_iterator operator ++(int) {
        zorder_iterator tmp = *this;
        ++(*this);
        return tmp;
    }

    /// dereference operator. provides access to the current window.
    HWND operator*() const { return _hwnd; }

    /// compare two iterators for equality.
    friend bool operator== (const zorder_iterator& a, const zorder_iterator& b) { return a._hwnd == b._hwnd && a._dir == b._dir; };

    /// compare two iterators for inequality.
    friend bool operator!= (const zorder_iterator& a, const zorder_iterator& b) { return a._hwnd != b._hwnd || a._dir != b._dir; };
};


/// a z-order iterator sentinel.
struct zorder_sentinel {};

/// compare a z-order iterator with a sentinel to determine if the iterator has reached the end.
inline bool operator ==(zorder_iterator it, zorder_sentinel) { return *it == NULL; }


static_assert(std::input_iterator<zorder_iterator>); // sanity check. zorder_iterator satifies the input iterator concept.
static_assert(std::sentinel_for<zorder_sentinel, zorder_iterator>); // sanity check. zorder_iterator has an associated sentinel.